#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "glcd.h"
#include "single_wire_UART.h"

void glcd_hardware_reset() {
	PORTB |= 0b00001000;
	_delay_ms(50);
	PORTB &= 0b11110111;
	_delay_ms(50);
	PORTB |= 0b00001000;
	_delay_ms(50);
}


void glcd_clear() {
	unsigned char i;
 	for (i=0;i<GLCD_WIDTH;i++) { vram[i] = 0; }  
}

void glcd_repaint() {
	unsigned char x,y;
	for (y=0;y<4;y++) {
		glcd_prepare_command();
		spi(0b10110000|y); //Set y Page (0-3)
		spi(0b00010000);   //set column 0, always 2byte
		spi(0b00000000);   //set column 0, always 2byte
		glcd_prepare_data();
		//_delay_us(100);
		for (x=0;x<GLCD_WIDTH;x++) { spi(vram[x]>>(y*8)); }
	}
}

void glcd_setpixel(unsigned char x, unsigned char y) {
	vram[x] |= ((GLCD_VRAM_TYPE) 1<<y);
}

void glcd_linehorizontal(unsigned char x1, unsigned char y1, unsigned char x2) {
	unsigned char i;
	if (x1>x2) {
		i = x2;
		x2 = x1;
		x1 = i;
	}
	for (i=x1;i<(x2+1);i++) { vram[i] |= ((GLCD_VRAM_TYPE) 1<<y1); }
}

void glcd_linevertical(unsigned char x1, unsigned char y1, unsigned char y2) {
  unsigned char i;
  GLCD_VRAM_TYPE temp=1;
  for (i=0;i<abs((int)y2-(int)y1);i++) {
    temp <<= 1;
    temp |= 1;
  }
  vram[x1] |= temp << y1;
}

void glcd_rectangle(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2) {
  glcd_linehorizontal(x1,y1,x2);
  glcd_linehorizontal(x1,y2,x2);
  glcd_linevertical(x1,y1,y2);
  glcd_linevertical(x2,y1,y2);
}

void glcd_line(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2) {
  signed int dx, dy;
  signed int stepx, stepy, fraction;
  dy = y2 - y1;
  dx = x2 - x1;

  if (dy < 0) {
    dy = -dy;
    stepy = -1;
  } else {
    stepy = 1;
  }
  
  if (dx < 0) {
    dx = -dx;
    stepx = -1;
  } else {
    stepx = 1;
  }
  
  dy <<= 1;
  dx <<= 1;
  glcd_setpixel(x1,y1);
  if (dx > dy) {
    fraction = dy - (dx >> 1); 
    while (x1 != x2) {
      if (fraction >= 0) {
        y1 += stepy;
        fraction -= dx;
      }
      x1 += stepx;
      fraction += dy;  
      glcd_setpixel(x1,y1);
    }
  } else {
    fraction = dx - (dy >> 1);
    while (y1 != y2) {
      if (fraction >= 0) {
        x1 += stepx;
        fraction -= dy;
      }
      y1 += stepy;
      fraction += dx;
      glcd_setpixel(x1,y1);
    }
  }
}


void glcd_contrast(unsigned char contrast) {
	glcd_prepare_command();
	spi(0b10000001); // Befehl fr "Kontrastwert setzen"
	spi(contrast);   // Und als Byte schreiben.
	glcd_prepare_data();
}

void glcd_puts(int x, int y, char *text, PGM_P font, char additionalSpacing, unsigned char color) {
	/*Fr APE 2.1 kompatible Fonts */
	unsigned int fontoffset, fontDataStart;
	unsigned char i, k, fontvbytes, charWidth, textPosition, fontStartByte, fontNumberOfChars;
	GLCD_VRAM_TYPE tempLong;
	int temp;


	// Zeichenhhe in Bytes (z.B. entpricht 14px genau 2 Bytes)
	// auch fr Abstand zwischen den Zeichen benutzt.
	fontvbytes        = (pgm_read_byte_near(&(font[3]))+7)/8;
	fontStartByte     = pgm_read_byte_near(&(font[4]));
	fontNumberOfChars = pgm_read_byte_near(&(font[5]));

	for (textPosition=0;textPosition<strlen(text);textPosition++) {
  		if(text[textPosition] < fontStartByte) continue; // Zeichennummer zu klein
		if(text[textPosition] > (fontStartByte+fontNumberOfChars) ) return; // Zeichennummer zu gro
		charWidth     = pgm_read_byte_near(&(font[   (PROP_FONT_HEADER_SIZE+text[textPosition]-fontStartByte)   ])); 
		fontDataStart = PROP_FONT_HEADER_SIZE + ((unsigned int) fontNumberOfChars); // Beginn der Fontdaten, da wir alle Width-Bytes noch berspringen mssen

		fontoffset = 0;

		// addiere alle Zeichenbreiten bis zum aktuellen Zeichen
		for (i=0; i<(text[textPosition] - pgm_read_byte(&(font[4]))); i++) { fontoffset += pgm_read_byte(&(font[PROP_FONT_HEADER_SIZE + i])); }
		fontoffset *= fontvbytes;    // alle davorliegenden Zeichen berspringen
		fontoffset += fontDataStart; // Nun haben wir die Startaddresse des Zeichens
		for (k=0;k<fontvbytes;k++) {
			//Fr jedes Zeilenbyte
			for (i=0;i<charWidth;i++) {
				//Fr korrektes Platzieren von negativen Koordinaten
				if (x+i<0) { continue; }

				// Fr jede Zeichenspalte
				// Fehler im FontCreator:
				// Bei mehrbytigen Schriftarten stimmt das letzte Byte nicht; es muss um 2 bit nach links geshifted werden zuerst,
				// daher gestaltet sich temp als Shiftoffset: y als koordinate + (Anzahl vertikale Bytes-1)*8 + (Schrifthhe in Pixel mod 8)
				if (k==(fontvbytes-1) && k<2) { temp = y + (k-1)*8  + (pgm_read_byte_near(&(font[3]))%8); } else { temp = y + (k*8); }

				// Teil der Schriftart lesen
				tempLong = pgm_read_byte_near(&(font[fontoffset+i+(k*charWidth)]));

				// Shift-Richtung bestimmen...
				if (temp < 0) {
					temp = abs(temp);
					if (color==COLOR_BLACK) { vram[x+i] |= tempLong >> temp; } else { vram[x+i] &= 0xFFFFFFFF & ~(tempLong >> temp); }
				} else {
					if (color==COLOR_BLACK) { vram[x+i] |= tempLong << temp; } else { vram[x+i] &= 0xFFFFFFFF & ~(tempLong << temp); }
				}
			}
		}
		x += charWidth;         // Abstand halten
		x += additionalSpacing; // Zusatzabstand halten, z.B. fr gesperrten Druck
		x++;                    // Eins dazu, ist der Miniabstand
		if (x>131) { break; }
	}
}


void glcd_image(int x, int y, PGM_P image) {
	//Zeichnet ein Image
	// 0: Breite in Pixeln
	// 1: Hhe in BYTE
	int width, height, colCounter;
	GLCD_VRAM_TYPE temp;

	width  = pgm_read_byte(&image[0]);
	height = pgm_read_byte(&image[1]);

	for (colCounter=0; colCounter<width; colCounter++) {
		// Hier htten wir nichts zu zeichnen,
		// da nicht im Display-Bereich
		if ( colCounter+x < 0 || colCounter+x > 131)  { continue; }

		temp  =  (unsigned long) pgm_read_byte_near(&image[colCounter]            +2);
		temp |= ((unsigned long) pgm_read_byte_near(&image[colCounter]+    width  +2) ) << 8;
		temp |= ((unsigned long) pgm_read_byte_near(&image[colCounter]+ (2*width) +2) ) << 16;
		temp |= ((unsigned long) pgm_read_byte_near(&image[colCounter]+ (3*width) +2) ) << 24;

		// Also runterschieben, da X + Zeile ein positiver Betrag ist
		if (y > 0) { temp >>= y; }
		// Also raufschieben, da X + Zeile ein negativer Betrag ist
		if (y < 0) { temp <<= abs(y); }
		vram[x + colCounter] |= temp;
	}
}

void glcd_filledrectangle(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char color) {
	unsigned char i;
	GLCD_VRAM_TYPE temp=1;
	for (i=0;i<abs((int)y2-(int)y1);i++) {
		temp <<= 1;
		temp |= 1;
	}
	temp <<= y1;             

	if (color == COLOR_BLACK) {
		for (i=x1;i<(x2+1);i++) { vram[i] |= temp; }
	} else {
		for (i=x1;i<(x2+1);i++) { vram[i] &= 0xFFFFFFFF & ~temp; }
	}
}



void spi(unsigned char myByte) {
	SPDR = myByte;
	while(!(SPSR & (1<<SPIF)));	
}
